{% extends "basic/layout.html" %}

{% block header %}
<div class="header">
 <a href="http://www.cubicweb.org">
  <img alt="cubicweb logo" src="{{ pathto('_static/cubicweb.png', 1) }}"/>
 </a>
</div>
{% endblock %}

{# puts the sidebar into "sidebar1" block i.e. before the document body #}
{% block sidebar1 %}{{ sidebar() }}{% endblock %}
{% block sidebar2 %}{% endblock %}
